/*   File: cZero.cc
     compute Zeros of holomorphic functions
    

     Author: Tomas Johnson <johnson@math.uu.se>      
     Latest edit: 
     By: Tomas Johnson <johnson@math.uu.se>
*/

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <stdexcept>
#include <time.h>
#include <unistd.h>
#include "list.h"

#include <cinterval.hpp>
#include "/mnt/calc/warwick/taylor/citaylor.hpp"

//#include "citaylor.hpp"

using namespace cxsc;
using namespace std;
using namespace taylor;

const interval PI(Pi());       // An enclosure of pi.
const cinterval I(interval(0), interval(1));
const int RECNUM(100);
enum command {START_TIMING, SHOW_TIMING, STOP_TIMING};


//////////////////////////////////////////////////////////
citaylor MyPow(const citaylor &z, int n) {
  if (n==2)
    return sqr(z);
  
  int k=get_order(z);
  
  if (k==0)
    return citaylor(0, power(get_j_coef(z,0),n));
  civector res(0,k);
  
  res[0]=power(get_j_coef(z,0),n);
  
  civector ztempV(0,k-1);
  for(int i=0; i<k; i++)
    ztempV[i]=get_j_coef(z,i);
  
  citaylor  ztemp(ztempV);

  citaylor tempf(MyPow(ztemp,n-1));
  citaylor dz=derivative(z, 1);
  citaylor gtemp=n*tempf*dz;
  for(int i=1; i<=k; i++)
    res[i]=get_j_coef(gtemp,i-1)/i;
  
  return citaylor(res);
}


/////////////////////////////////////////////////////
citaylor aPow(const citaylor &z, int n) {
  if ( !(0.0 <= get_j_coef(z,0)) && (diam(Re(get_j_coef(z,0)))>0.00001) ) {
    citaylor res=power(z,n);
    return res;
  }
  
  if (n==1)
    return z;
  if (n==2)
    return sqr(z);
  int k=n/2;
  if (n==2*k) 
    return sqr(aPow(z,k));
  else
    return z*sqr(aPow(z,k));
  
}

////////////////////////////////////////////////////////////////////////////
//
// 'START_TIMING' starts the clock, 'STOP_TIMING' stops the clock,
// and prints the elapsed time in hh:mm:ss. 'SHOW_TIMING' prints the
// elapsed time.
//
void clock(const command &cmd)
{
  struct tm     *ptr;          // Used for timing the program
  static time_t  time_start;   // start time

  if ( cmd == START_TIMING ) {
    time_start = time(NULL);
    ptr = localtime(&time_start);
    cout << "Started computations: " << asctime(ptr);
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;

    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
      cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: ";
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) { seconds -= 60; minutes++; }
    while ( minutes >= 60 ) { minutes -= 60;   hours++; }
    while (   hours >= 24 ) {   hours -= 24;    days++; }
    while (    days >=  7 ) {    days -=  7;   weeks++; }
    if ( weeks > 0)
      cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
      cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
      cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
      cout << minutes << " minute" << (minutes != 1 ? "s, " : ", ");
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl;
  }
}

////////////////////////////////////////////////////////////////////////////


citaylor zoft(int id, citaylor t, complex llC, complex urC) {
  if(id==1){
    complex startP(Re(urC) , Im(llC));
    complex stopP(urC);
    return startP + t*(stopP-startP);
  }
  if(id==2) {
    complex startP(urC);
    complex stopP(Re(llC), Im(urC));
    return startP + t*(stopP-startP);
  }
  if(id==3) {
    complex startP(Re(llC),Im(urC));
    complex stopP(llC);
    return startP + t*(stopP-startP);
  }
  if(id==4) {
    complex startP(llC);
    complex stopP(Re(urC), Im(llC));
    return startP + t*(stopP-startP);
  }
  cout<<"ERROR"<<endl;
  exit(0);
}

citaylor fcnF(citaylor z, bool &fOK) {
  // return aPow(z,11)-1;
  //  ENTER CONDITIONS ON f's domain!!
 //  if (0<=get_j_coef(z,0)) {
//     fOK=false;
//     return citaylor(get_order(z),1);
//   }
  //Herlocker, Ely:
  //return power(z,3) - sqr(z) - z - 2; //(1)
  //Some other examples
  //return power(z,3) - sqr(z) - z - 2+sin(10*sqr(z)); //(2)
  //return sin(z);  //(3)
  
  //Dellnitz, Schtze, Zheng
  //return power(z,50)+power(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)
  return aPow(z,50)+aPow(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)
  //return MyPow(z,50)+MyPow(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)
  //return sqr(z)+interval(-19435)/interval(pow(10,6))*z+interval(100041)/interval(100)*
  //exp(-interval(5)/interval(1000)*z)+interval(522463); //(5)

 //  interval A(166689);
//   A=A/interval(pow(10,7));
//   interval B(166659);
//   B=B/interval(pow(10,7));
//   interval C(-37485);
//   C=C/interval(pow(10,9));
//   interval D(166667);
//   D=D/interval(pow(10,7));
  // interval t(2.0);
//   return ( ((A+interval(212)/interval(pow(10,16))*z)*(D+interval(6.0)/interval(pow(10,16))*z))-
// 	   ((B+exp(-t*z)*(C+interval(6.0)/interval(pow(10,16))*z))*
// 	    (interval(1.0)/interval(60.0)+interval(6.0)/interval(pow(10,16))*exp(-t*z)*z)) ); //(6)

  //return interval(70)*sqr(sqr(z)-2*z+3)*(z-1-interval(99)/interval(70)*I); //(7)

}

citaylor gFcn(int id, citaylor t, complex llc, complex urC, bool &ok, int order) {
  if(ok) {
    citaylor z=zoft(id, t, llc, urC);
    bool fOK(true);
    
    citaylor f=fcnF(z, fOK);
    if (!fOK || (0<=get_j_coef(f,order-1)) ) {
      ok=false; return citaylor((get_order(t)-order),1); }
    
    citaylor df=derivative(f, order);
    citaylor ff;
    if (order>1)
      ff=derivative(f, order-1);
    else 
      ff=f;
    
    int k=get_order(ff);
    civector fcoeff=get_all_coef(ff);
    civector fTcoeff(0,k-1);
    for(int i=0;i<k;i++)
      fTcoeff[i]=fcoeff[i];
    
    citaylor fT(fTcoeff);
    
    return df/fT;
  }
  else
    return citaylor((get_order(t)-order),1);
}


interval simInt(int id, interval t, complex llC, complex urC, bool &ok, int order) {
  citaylor tl(order, Inf(t));
  citaylor tm(order, mid(t));
  citaylor tu(order, Sup(t));
  citaylor approx=(diam(t))/6*(gFcn(id, tl, llC, urC, ok, order)+4*gFcn(id, tm, llC, urC, ok, order)
			       +gFcn(id, tu, llC, urC, ok, order));
  if (!ok)
    return t;
  citaylor terr(4+order,t);
  cinterval err=-get_j_coef(gFcn(id, terr, llC, urC, ok, order),4)*pow(diam(t)/2,5)*4/15;
  if (!ok)
    return t;
  interval imRes=Im(get_j_coef(approx,0)+err);
  
  return imRes/(2*PI);
}

//////////////////////////////////////////////////////////////////////////////////////////////
//fail =-1, possible zero on the contour
//fail = 0, to many iterations
//fail = 1, OK

interval quadSide(int id, interval t, complex llC, complex urC, real tol, int &fail, int rec, int order) {
  if (fail==-1)
    return t;
  
  if (fail==0 || (rec>RECNUM) ) {
    fail=0;
    return t;
  } 
    
  bool ok(true);
  interval res=simInt(id, t, llC, urC, ok, order);
  
  if (!ok && rec>=RECNUM) {
    fail=-1;
    return t;
  }
  
  if(!ok || diam(res)>tol) {
    res=quadSide(id, interval(Inf(t),mid(t)), llC, urC, tol/2, fail, rec+1, order)
      +quadSide(id, interval(mid(t), Sup(t)), llC, urC, tol/2, fail, rec+1, order); 
  }
  return res;
}

interval quad(cinterval Dom, int &fail, int order) {
  interval res(0);
  complex urC(Sup(Re(Dom)), Sup(Im(Dom)));
  complex llC(Inf(Re(Dom)), Inf(Im(Dom)));
  real tol(0.25);
  for(int id=1; id<=4; id++)
    res+=quadSide(id, interval(0,1), llC, urC, tol, fail, 0, order);
  return res;
}

////////////////////////////////////////////////////////////////////////////
//
// Bisects in both components, and stores all four squares.
//
void splitAndStore(const cinterval &z, List<cinterval> &workList)
{
  interval reZ = Re(z);
  interval imZ = Im(z);

  workList+= cinterval(interval(Inf(reZ), mid(reZ)), interval(Inf(imZ), mid(imZ)));
  workList+= cinterval(interval(Inf(reZ), mid(reZ)), interval(mid(imZ), Sup(imZ)));
  workList+= cinterval(interval(mid(reZ), Sup(reZ)), interval(Inf(imZ), mid(imZ)));
  workList+= cinterval(interval(mid(reZ), Sup(reZ)), interval(mid(imZ), Sup(imZ)));
}

////////////////////////////////////////////////////////////////////////////

bool searchZero(cinterval &Dom, int NZ, int ntolp) {
  try {
    real eps(pow(10,-ntolp));
    citaylor z(1,mid(Dom));
    bool fOK(true);
    citaylor f=fcnF(z,fOK);
    if (!fOK)
      return false;
    
    complex zz;
    if (0.0 == mid(get_j_coef(f,1)) )
      return false;
    else
      zz=mid(Dom)-mid(get_j_coef(f,0))/mid(get_j_coef(f,1));
    if (!(zz<=Dom))
      return false;
    
    int i=0;
    int nits=max(5,ntolp);
    while(i<nits && zz<=Dom) {
      citaylor zt(1,zz);
      f=fcnF(zt, fOK);
      if (!fOK || (0.0==mid(get_j_coef(f,1))) ){
	i=nits;
      }
      else {
	complex zztemp=zz-mid(get_j_coef(f,0))/mid(get_j_coef(f,1));
	if(zztemp<=Dom) {
	  zz=zztemp;
	  i=i+1;
	}
	else{
	  i=nits;
	}
      }
    }
    int fail(1);
    cinterval Dom2;
    interval arg;
    Dom2=(zz+cinterval(interval(-eps,eps),interval(-eps,eps))) & Dom;
    arg=quad(Dom2, fail, 1);
    if(fail!=1)
      return false;
    
    if ((ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) && (floor(_double(Sup(arg)))==NZ) ) {
      Dom=Dom2;
      return true;
    }
  }
  catch(...){}
  return false;
}

bool MultipleRoot(cinterval Dom, int NZ) {
  int fail(1);
  interval arg=quad(Dom, fail, NZ);
  if(fail!=1)
    return 1;
  if (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) {
    int resNZ=(int)ceil(_double(Inf(arg)));
    if(resNZ!=0)
      return 1;
    else
      return 0;
  }
  else
    return 1;
  
}

int main(int argc, char * argv[])
{
  if ( argc != 8 ) {
    cout << "Syntax: " << argv[0] << " inf(real) sup(real) inf(im) sup(im) sTol nTol mRoot" << endl<<endl
	 << "Search box is: [inf(real), sup(real)]+i[inf(im), sup(im)]"<<endl
	 << "sTol is the splitting tolerance, i.e the smallest diameter a box is split into."<<endl
	 << "10^{-nTol}, is the inflation box used for the result of Newtons method."<<endl
	 << "mRoot : check if clusters do not contain multiple roots, (slow if sTol large). 1=on, 0=off"<<endl;
    exit(0);
  }
  cout << endl
       << "**********************************************************" << endl
       << "************* cZero - complex root finder** **************" << endl
       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__
       << endl << endl;
  
  interval realDom(atof(argv[1]),atof(argv[2]) );
  interval imDom(atof(argv[3]),atof(argv[4]) );
  cinterval Dom(realDom, imDom);
  real stol(atof(argv[5]));
  int ntolp(atoi(argv[6]));
  int mRoot(atoi(argv[7]));
  cout<< SetPrecision(17,16);
  List<cinterval> workList, tempList;

  clock(START_TIMING);
  
  workList+=Dom;
  while(!IsEmpty(workList)) {
    cinterval workDom=Pop(workList);
    //cout<<workDom<<endl;
    int fail(1);
    interval arg(0);
    try{
      arg=quad(workDom, fail, 1);
    }
    catch(...){fail=0;}

    if(fail==-1){
      cout<<"zero on contour, inflating!"<<endl;
      workList+=cinterval(interval(Inf(Re(workDom))-stol/2, Sup(Re(workDom))+stol/2) , 
			 interval(Inf(Im(workDom))-stol/2, Sup(Im(workDom))+stol/2) );
    }
    else if(fail==0) {
      if (diam(abs(workDom))<stol)
	cout<<"cannot integrate on : " <<Re(workDom)<<" + i"<<Im(workDom)<<endl;
      else
	splitAndStore(workDom, workList); 
    }
    else {
      if (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) {
	int NZ=(int)ceil(_double(Inf(arg)));
	
	if(workDom==Dom)
	  cout<<"Total number of zeros = "<<NZ<<endl;
	if (NZ==1) {
	  if(searchZero(workDom,1, ntolp) || (diam(abs(workDom))<stol))
	    cout<<"There is: "<< NZ << " zero in the domain " <<Re(workDom)<<" + i"<<Im(workDom)<<endl;
	  else
	    splitAndStore(workDom, workList);
	}
	else if(NZ>1) {
	  if (diam(abs(workDom))<stol) {
	    cout<<"There are: "<< NZ << " zeros in the domain " <<Re(workDom)<<" + i"<<Im(workDom)<<endl; 
	    if(mRoot==1) {
	      while(NZ>1 && !MultipleRoot(workDom, NZ)) {
		cout<<"There are no zeros of order "<< NZ <<" in the domain"<<endl;
		NZ--;
	      }
	    }
	  }
	  else {
	    splitAndStore(workDom, workList);}
	}
      }
    }
  }
  clock(STOP_TIMING);
  return 0;
}

////////////////////////////////////////////////////////////////////////////
